"""
    tests.wsgi
    ~~~~~~~~~~

    Tests the WSGI utilities.

    :copyright:
        (c) 2014 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest
from werkzeug.test import Client, create_environ, run_wsgi_app

from os import path
from io import StringIO, BytesIO
from tempfile import TemporaryDirectory
from contextlib import closing

from verktyg.responses import BaseResponse
from verktyg.exceptions import BadRequest, ClientDisconnected
from verktyg import wsgi
from werkzeug._compat import to_bytes


class WsgiTestCase(unittest.TestCase):
    def test_shareddatamiddleware_get_file_loader(self):
        app = wsgi.SharedDataMiddleware(None, {})
        self.assertTrue(callable(app.get_file_loader('foo')))

    def test_shared_data_middleware(self):
        def null_application(environ, start_response):
            start_response('404 NOT FOUND', [('Content-Type', 'text/plain')])
            yield b'NOT FOUND'

        with TemporaryDirectory() as test_dir:
            test_file_name = path.join(test_dir, 'äöü').encode()
            with open(test_file_name, 'w') as test_file:
                test_file.write('FOUND')

            app = wsgi.SharedDataMiddleware(null_application, {
                '/':        path.join(path.dirname(__file__), 'resources'),
                '/sources': path.join(path.dirname(__file__), 'resources'),
                '/pkg':     ('werkzeug.debug', 'shared'),
                '/foo':     test_dir
            })

            for p in '/test.txt', '/sources/test.txt', '/foo/äöü':
                with self.subTest(path=p):
                    app_iter, status, headers = run_wsgi_app(
                        app, create_environ(p)
                    )
                    self.assertEqual(status, '200 OK')
                    with closing(app_iter) as app_iter:
                        data = b''.join(app_iter).strip()
                    self.assertEqual(data, b'FOUND')

            app_iter, status, headers = run_wsgi_app(
                app, create_environ('/pkg/debugger.js'))
            with closing(app_iter) as app_iter:
                contents = b''.join(app_iter)
            self.assertIn(b'$(function() {', contents)

            app_iter, status, headers = run_wsgi_app(
                app, create_environ('/missing'))
            self.assertEqual(status, '404 NOT FOUND')
            self.assertEqual(b''.join(app_iter).strip(), b'NOT FOUND')

    def test_dispatchermiddleware(self):
        def null_application(environ, start_response):
            start_response('404 NOT FOUND', [('Content-Type', 'text/plain')])
            yield b'NOT FOUND'

        def dummy_application(environ, start_response):
            start_response('200 OK', [('Content-Type', 'text/plain')])
            yield to_bytes(environ['SCRIPT_NAME'])

        app = wsgi.DispatcherMiddleware(null_application, {
            '/test1': dummy_application,
            '/test2/very': dummy_application,
        })
        tests = {
            '/test1': ('/test1', '/test1/asfd', '/test1/very'),
            '/test2/very': (
                '/test2/very', '/test2/very/long/path/after/script/name'
            ),
        }
        for name, urls in tests.items():
            for p in urls:
                environ = create_environ(p)
                app_iter, status, headers = run_wsgi_app(app, environ)
                self.assertEqual(status, '200 OK')
                self.assertEqual(b''.join(app_iter).strip(), to_bytes(name))

        app_iter, status, headers = run_wsgi_app(
            app, create_environ('/missing'))
        self.assertEqual(status, '404 NOT FOUND')
        self.assertEqual(b''.join(app_iter).strip(), b'NOT FOUND')

    def test_get_host(self):
        env = {
            'HTTP_X_FORWARDED_HOST': 'example.org',
            'SERVER_NAME': 'bullshit', 'HOST_NAME': 'ignore me dammit',
        }
        self.assertEqual(wsgi.get_host(env), 'example.org')
        self.assertEqual(
            wsgi.get_host(create_environ('/', 'http://example.org')),
            'example.org'
        )

    def test_get_host_multiple_forwarded(self):
        env = {
            'HTTP_X_FORWARDED_HOST': 'example.com, example.org',
            'SERVER_NAME': 'bullshit', 'HOST_NAME': 'ignore me dammit',
        }
        self.assertEqual(wsgi.get_host(env), 'example.com')
        self.assertEqual(
            wsgi.get_host(create_environ('/', 'http://example.com')),
            'example.com'
        )

    def test_get_host_validation(self):
        env = {
            'HTTP_X_FORWARDED_HOST': 'example.org',
            'SERVER_NAME': 'bullshit', 'HOST_NAME': 'ignore me dammit',
        }
        self.assertEqual(
            wsgi.get_host(env, trusted_hosts=['.example.org']), 'example.org'
        )
        self.assertRaises(
            BadRequest, wsgi.get_host, env, trusted_hosts=['example.com']
        )

    def test_responder(self):
        def foo(environ, start_response):
            return BaseResponse(b'Test')
        client = Client(wsgi.responder(foo), BaseResponse)
        response = client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_data(), b'Test')

    def test_pop_path_info(self):
        original_env = {'SCRIPT_NAME': '/foo', 'PATH_INFO': '/a/b///c'}

        # regular path info popping
        def assert_tuple(script_name, path_info):
            self.assertEqual(env.get('SCRIPT_NAME'), script_name)
            self.assertEqual(env.get('PATH_INFO'), path_info)
        env = original_env.copy()

        def pop():
            return wsgi.pop_path_info(env)

        assert_tuple('/foo', '/a/b///c')
        self.assertEqual(pop(), 'a')
        assert_tuple('/foo/a', '/b///c')
        self.assertEqual(pop(), 'b')
        assert_tuple('/foo/a/b', '///c')
        self.assertEqual(pop(), 'c')
        assert_tuple('/foo/a/b///c', '')
        self.assertIsNone(pop())

    def test_peek_path_info(self):
        env = {
            'SCRIPT_NAME': '/foo',
            'PATH_INFO': '/aaa/b///c'
        }

        self.assertEqual(wsgi.peek_path_info(env), 'aaa')
        self.assertEqual(wsgi.peek_path_info(env), 'aaa')
        self.assertEqual(wsgi.peek_path_info(env, charset=None), b'aaa')
        self.assertEqual(wsgi.peek_path_info(env, charset=None), b'aaa')

    def test_path_info_and_script_name_fetching(self):
        env = create_environ('/\N{SNOWMAN}', 'http://example.com/\N{COMET}/')
        self.assertEqual(
            wsgi.get_path_info(env),
            '/\N{SNOWMAN}'
        )
        self.assertEqual(
            wsgi.get_path_info(env, charset=None),
            '/\N{SNOWMAN}'.encode('utf-8')
        )
        self.assertEqual(
            wsgi.get_script_name(env),
            '/\N{COMET}'
        )
        self.assertEqual(
            wsgi.get_script_name(env, charset=None),
            '/\N{COMET}'.encode('utf-8')
        )

    def test_query_string_fetching(self):
        env = create_environ('/?\N{SNOWMAN}=\N{COMET}')
        qs = wsgi.get_query_string(env)
        self.assertEqual(qs, '%E2%98%83=%E2%98%84')

    def test_limited_stream(self):
        class RaisingLimitedStream(wsgi.LimitedStream):
            def on_exhausted(self):
                raise BadRequest('input stream exhausted')

        io = BytesIO(b'123456')
        stream = RaisingLimitedStream(io, 3)
        self.assertEqual(stream.read(), b'123')
        self.assertRaises(BadRequest, stream.read)

        io = BytesIO(b'123456')
        stream = RaisingLimitedStream(io, 3)
        self.assertEqual(stream.tell(), 0)
        self.assertEqual(stream.read(1), b'1')
        self.assertEqual(stream.tell(), 1)
        self.assertEqual(stream.read(1), b'2')
        self.assertEqual(stream.tell(), 2)
        self.assertEqual(stream.read(1), b'3')
        self.assertEqual(stream.tell(), 3)
        self.assertRaises(BadRequest, stream.read)

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readline(), b'123456\n')
        self.assertEqual(stream.readline(), b'ab')

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readlines(), [b'123456\n', b'ab'])

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readlines(2), [b'12'])
        self.assertEqual(stream.readlines(2), [b'34'])
        self.assertEqual(stream.readlines(), [b'56\n', b'ab'])

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readline(100), b'123456\n')

        io = BytesIO(b'123456\nabcdefg')
        stream = wsgi.LimitedStream(io, 9)
        self.assertEqual(stream.readlines(100), [b'123456\n', b'ab'])

        io = BytesIO(b'123456')
        stream = wsgi.LimitedStream(io, 3)
        self.assertEqual(stream.read(1), b'1')
        self.assertEqual(stream.read(1), b'2')
        self.assertEqual(stream.read(), b'3')
        self.assertEqual(stream.read(), b'')

        io = BytesIO(b'123456')
        stream = wsgi.LimitedStream(io, 3)
        self.assertEqual(stream.read(-1), b'123')

        io = BytesIO(b'123456')
        stream = wsgi.LimitedStream(io, 0)
        self.assertEqual(stream.read(-1), b'')

        io = StringIO('123456')
        stream = wsgi.LimitedStream(io, 0)
        self.assertEqual(stream.read(-1), '')

        io = StringIO('123\n456\n')
        stream = wsgi.LimitedStream(io, 8)
        self.assertEqual(list(stream), ['123\n', '456\n'])

    def test_limited_stream_disconnection(self):
        io = BytesIO(b'A bit of content')

        # disconnect detection on out of bytes
        stream = wsgi.LimitedStream(io, 255)
        self.assertRaises(ClientDisconnected, stream.read)

        # disconnect detection because file close
        io = BytesIO(b'x' * 255)
        io.close()
        stream = wsgi.LimitedStream(io, 255)
        self.assertRaises(ClientDisconnected, stream.read)

    def test_path_info_extraction(self):
        x = wsgi.extract_path_info('http://example.com/app', '/app/hello')
        self.assertEqual(x, '/hello')
        x = wsgi.extract_path_info(
            'http://example.com/app', 'https://example.com/app/hello'
        )
        self.assertEqual(x, '/hello')
        x = wsgi.extract_path_info(
            'http://example.com/app/', 'https://example.com/app/hello'
        )
        self.assertEqual(x, '/hello')
        x = wsgi.extract_path_info('http://example.com/app/',
                                   'https://example.com/app')
        self.assertEqual(x, '/')
        x = wsgi.extract_path_info('http://☃.net/', '/fööbär')
        self.assertEqual(x, '/fööbär')
        x = wsgi.extract_path_info('http://☃.net/x', 'http://☃.net/x/fööbär')
        self.assertEqual(x, '/fööbär')

        env = create_environ('/fööbär', 'http://☃.net/x/')
        x = wsgi.extract_path_info(env, 'http://☃.net/x/fööbär')
        self.assertEqual(x, '/fööbär')

        x = wsgi.extract_path_info('http://example.com/app/',
                                   'https://example.com/a/hello')
        self.assertIsNone(x)
        x = wsgi.extract_path_info('http://example.com/app/',
                                   'https://example.com/app/hello',
                                   collapse_http_schemes=False)
        self.assertIsNone(x)

    def test_get_host_fallback(self):
        self.assertEqual(
            wsgi.get_host({
                'SERVER_NAME':      'foobar.example.com',
                'wsgi.url_scheme':  'http',
                'SERVER_PORT':      '80'
            }),
            'foobar.example.com'
        )
        self.assertEqual(
            wsgi.get_host({
                'SERVER_NAME':      'foobar.example.com',
                'wsgi.url_scheme':  'http',
                'SERVER_PORT':      '81'
            }),
            'foobar.example.com:81'
        )

    def test_get_current_url_unicode(self):
        env = create_environ()
        env['QUERY_STRING'] = 'foo=bar&baz=blah&meh=\xcf'
        rv = wsgi.get_current_url(env)
        self.assertEqual(
            rv, 'http://localhost/?foo=bar&baz=blah&meh=\ufffd'
        )

    def test_multi_part_line_breaks(self):
        data = 'abcdef\r\nghijkl\r\nmnopqrstuvwxyz\r\nABCDEFGHIJK'
        test_stream = StringIO(data)
        lines = list(wsgi.make_line_iter(test_stream, limit=len(data),
                                         buffer_size=16))
        self.assertEqual(
            lines, [
                'abcdef\r\n', 'ghijkl\r\n', 'mnopqrstuvwxyz\r\n', 'ABCDEFGHIJK'
            ]
        )

        data = 'abc\r\nThis line is broken by the buffer length.' \
            '\r\nFoo bar baz'
        test_stream = StringIO(data)
        lines = list(wsgi.make_line_iter(test_stream, limit=len(data),
                                         buffer_size=24))
        self.assertEqual(
            lines, [
                'abc\r\n',
                'This line is broken by the buffer length.\r\n',
                'Foo bar baz',
            ]
        )

    def test_multi_part_line_breaks_bytes(self):
        data = b'abcdef\r\nghijkl\r\nmnopqrstuvwxyz\r\nABCDEFGHIJK'
        test_stream = BytesIO(data)
        lines = list(wsgi.make_line_iter(test_stream, limit=len(data),
                                         buffer_size=16))
        self.assertEqual(
            lines, [
                b'abcdef\r\n', b'ghijkl\r\n',
                b'mnopqrstuvwxyz\r\n', b'ABCDEFGHIJK'
            ]
        )

        data = b'abc\r\nThis line is broken by the buffer length.' \
            b'\r\nFoo bar baz'
        test_stream = BytesIO(data)
        lines = list(wsgi.make_line_iter(test_stream, limit=len(data),
                                         buffer_size=24))
        self.assertEqual(
            lines, [
                b'abc\r\n',
                b'This line is broken by the buffer length.\r\n',
                b'Foo bar baz',
            ]
        )

    def test_multi_part_line_breaks_problematic(self):
        data = 'abc\rdef\r\nghi'
        for x in range(1, 10):
            test_stream = StringIO(data)
            lines = list(wsgi.make_line_iter(test_stream, limit=len(data),
                                             buffer_size=4))
            self.assertEqual(lines, ['abc\r', 'def\r\n', 'ghi'])

    def test_iter_functions_support_iterators(self):
        data = ['abcdef\r\nghi', 'jkl\r\nmnopqrstuvwxyz\r', '\nABCDEFGHIJK']
        lines = list(wsgi.make_line_iter(data))
        self.assertEqual(
            lines, [
                'abcdef\r\n', 'ghijkl\r\n', 'mnopqrstuvwxyz\r\n', 'ABCDEFGHIJK'
            ]
        )

    def test_make_chunk_iter(self):
        data = ['abcdefXghi', 'jklXmnopqrstuvwxyzX', 'ABCDEFGHIJK']
        rv = list(wsgi.make_chunk_iter(data, 'X'))
        self.assertEqual(
            rv, ['abcdef', 'ghijkl', 'mnopqrstuvwxyz', 'ABCDEFGHIJK']
        )

        data = 'abcdefXghijklXmnopqrstuvwxyzXABCDEFGHIJK'
        test_stream = StringIO(data)
        rv = list(wsgi.make_chunk_iter(test_stream, 'X', limit=len(data),
                                       buffer_size=4))
        self.assertEqual(
            rv, ['abcdef', 'ghijkl', 'mnopqrstuvwxyz', 'ABCDEFGHIJK']
        )

    def test_make_chunk_iter_bytes(self):
        data = [b'abcdefXghi', b'jklXmnopqrstuvwxyzX', b'ABCDEFGHIJK']
        rv = list(wsgi.make_chunk_iter(data, 'X'))
        self.assertEqual(
            rv, [b'abcdef', b'ghijkl', b'mnopqrstuvwxyz', b'ABCDEFGHIJK']
        )

        data = b'abcdefXghijklXmnopqrstuvwxyzXABCDEFGHIJK'
        test_stream = BytesIO(data)
        rv = list(wsgi.make_chunk_iter(test_stream, 'X', limit=len(data),
                                       buffer_size=4))
        self.assertEqual(
            rv, [b'abcdef', b'ghijkl', b'mnopqrstuvwxyz', b'ABCDEFGHIJK']
        )

    def test_lines_longer_buffer_size(self):
        data = '1234567890\n1234567890\n'
        for bufsize in range(1, 15):
            lines = list(wsgi.make_line_iter(
                StringIO(data), limit=len(data), buffer_size=4
            ))
            self.assertEqual(lines, ['1234567890\n', '1234567890\n'])



"""
    verktyg.testsuite
    ~~~~~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from verktyg.testsuite import (
    test_accept, test_application, test_dispatch, test_exceptions,
    test_routing, test_views, test_requests, test_responses, test_wsgi,
)


loader = unittest.TestLoader()
suite = unittest.TestSuite((
    loader.loadTestsFromModule(test_accept),
    loader.loadTestsFromModule(test_application),
    loader.loadTestsFromModule(test_dispatch),
    loader.loadTestsFromModule(test_exceptions),
    loader.loadTestsFromModule(test_routing),
    loader.loadTestsFromModule(test_views),
    loader.loadTestsFromModule(test_requests),
    loader.loadTestsFromModule(test_responses),
    loader.loadTestsFromModule(test_wsgi),
))



"""
    verktyg.jinja2
    ~~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""


def add_templates(self, *loaders):
    for loader in loaders:
        self.jinja_env.loader.loaders.append(loader)


def get_renderer(self, name):
    if isinstance(name, str):
        return self.jinja_env.get_template(name).render
    else:
        # assume callable
        return name


def bind(app, *loaders, **kwargs):
    """ Add a jinja2 environment to an application
    """
    # imported here as jinja is not required in setup.py.
    import jinja2
    app.jinja_env = jinja2.Environment(
        loader=jinja2.ChoiceLoader([]), **kwargs
    )

    app.add_method('add_templates', add_templates)
    app.add_method('get_renderer', get_renderer)

    app.add_templates(*loaders)



"""
    tests.requests
    ~~~~~~~~~~~~~~

    Tests for request objects.

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

import pickle
from io import BytesIO
from datetime import datetime

from werkzeug.datastructures import (
    MultiDict, ImmutableOrderedMultiDict, ImmutableList,
    ImmutableTypeConversionDict, CombinedMultiDict,
    CharsetAccept, MIMEAccept, LanguageAccept, Accept,
)
from werkzeug.test import Client

from verktyg.wsgi import LimitedStream
from verktyg.exceptions import SecurityError
from verktyg.requests import BaseRequest, Request, PlainRequest
from verktyg.responses import BaseResponse


class RequestTestResponse(BaseResponse):

    """Subclass of the normal response class we use to test response
    and base classes.  Has some methods to test if things in the
    response match.
    """

    def __init__(self, response, status, headers):
        BaseResponse.__init__(self, response, status, headers)
        self.body_data = pickle.loads(self.get_data())

    def __getitem__(self, key):
        return self.body_data[key]


def request_demo_app(environ, start_response):
    request = BaseRequest(environ)
    assert 'verktyg.request' in environ
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return [pickle.dumps({
        'args':             request.args,
        'args_as_list':     list(request.args.lists()),
        'form':             request.form,
        'form_as_list':     list(request.form.lists()),
        'environ':          prepare_environ_pickle(request.environ),
        'data':             request.get_data()
    })]


def prepare_environ_pickle(environ):
    result = {}
    for key, value in environ.items():
        try:
            pickle.dumps((key, value))
        except Exception:
            continue
        result[key] = value
    return result


class RequestsTestCase(unittest.TestCase):
    def assert_environ(self, environ, method):
        self.assertEqual(environ['REQUEST_METHOD'], method)
        self.assertEqual(environ['PATH_INFO'], '/')
        self.assertEqual(environ['SCRIPT_NAME'], '')
        self.assertEqual(environ['SERVER_NAME'], 'localhost')
        self.assertEqual(environ['wsgi.version'], (1, 0))
        self.assertEqual(environ['wsgi.url_scheme'], 'http')

    def test_base_request(self):
        client = Client(request_demo_app, RequestTestResponse)

        # get requests
        response = client.get('/?foo=bar&foo=hehe')
        self.assertEqual(
            response['args'], MultiDict([('foo', u'bar'), ('foo', u'hehe')])
        )
        self.assertEqual(
            response['args_as_list'], [('foo', [u'bar', u'hehe'])]
        )
        self.assertEqual(response['form'], MultiDict())
        self.assertEqual(response['form_as_list'], [])
        self.assertEqual(response['data'], b'')
        self.assert_environ(response['environ'], 'GET')

        # post requests with form data
        response = client.post(
            '/?blub=blah', data='foo=blub+hehe&blah=42',
            content_type='application/x-www-form-urlencoded'
        )
        self.assertEqual(response['args'], MultiDict([('blub', u'blah')]))
        self.assertEqual(response['args_as_list'], [('blub', [u'blah'])])
        self.assertEqual(
            response['form'], MultiDict([
                ('foo', u'blub hehe'), ('blah', u'42'),
            ])
        )
        self.assertEqual(response['data'], b'')
        self.assert_environ(response['environ'], 'POST')

        # patch requests with form data
        response = client.patch(
            '/?blub=blah', data='foo=blub+hehe&blah=42',
            content_type='application/x-www-form-urlencoded'
        )
        self.assertEqual(response['args'], MultiDict([('blub', u'blah')]))
        self.assertEqual(response['args_as_list'], [('blub', [u'blah'])])
        self.assertEqual(
            response['form'], MultiDict([
                ('foo', u'blub hehe'), ('blah', u'42'),
            ])
        )
        self.assertEqual(response['data'], b'')
        self.assert_environ(response['environ'], 'PATCH')

        # post requests with json data
        json = b'{"foo": "bar", "blub": "blah"}'
        response = client.post(
            '/?a=b', data=json, content_type='application/json'
        )
        self.assertEqual(response['data'], json)
        self.assertEqual(response['args'], MultiDict([('a', u'b')]))
        self.assertEqual(response['form'], MultiDict())

    def test_query_string_is_bytes(self):
        req = Request.from_values(u'/?foo=%2f')
        self.assertEqual(req.query_string, b'foo=%2f')

    def test_request_repr(self):
        req = Request.from_values('/foobar')
        self.assertEqual(
            "<Request 'http://localhost/foobar' [GET]>", repr(req)
        )
        # test with non-ascii characters
        req = Request.from_values('/привет')
        self.assertEqual(
            "<Request 'http://localhost/привет' [GET]>", repr(req)
        )
        # test with unicode type for python 2
        req = Request.from_values(u'/привет')
        self.assertEqual(
            "<Request 'http://localhost/привет' [GET]>", repr(req)
        )

    def test_access_route(self):
        req = Request.from_values(headers={
            'X-Forwarded-For': '192.168.1.2, 192.168.1.1'
        })
        req.environ['REMOTE_ADDR'] = '192.168.1.3'
        self.assertEqual(req.access_route, ['192.168.1.2', '192.168.1.1'])
        self.assertEqual(req.remote_addr, '192.168.1.3')

        req = Request.from_values()
        req.environ['REMOTE_ADDR'] = '192.168.1.3'
        self.assertEqual(list(req.access_route), ['192.168.1.3'])

    def test_url_request_descriptors(self):
        req = Request.from_values(
            '/bar?foo=baz', 'http://example.com/test'
        )
        self.assertEqual(req.path, u'/bar')
        self.assertEqual(req.full_path, u'/bar?foo=baz')
        self.assertEqual(req.script_root, u'/test')
        self.assertEqual(req.url, u'http://example.com/test/bar?foo=baz')
        self.assertEqual(req.base_url, u'http://example.com/test/bar')
        self.assertEqual(req.url_root, u'http://example.com/test/')
        self.assertEqual(req.host_url, u'http://example.com/')
        self.assertEqual(req.host, 'example.com')
        self.assertEqual(req.scheme, 'http')

        req = Request.from_values(
            '/bar?foo=baz', 'https://example.com/test'
        )
        self.assertEqual(req.scheme, 'https')

    def test_url_request_descriptors_query_quoting(self):
        next = 'http%3A%2F%2Fwww.example.com%2F%3Fnext%3D%2F'
        req = Request.from_values(
            '/bar?next=' + next, 'http://example.com/'
        )
        self.assertEqual(req.path, u'/bar')
        self.assertEqual(req.full_path, u'/bar?next=' + next)
        self.assertEqual(req.url, u'http://example.com/bar?next=' + next)

    def test_url_request_descriptors_hosts(self):
        req = Request.from_values(
            '/bar?foo=baz', 'http://example.com/test'
        )
        req.trusted_hosts = ['example.com']
        self.assertEqual(req.path, u'/bar')
        self.assertEqual(req.full_path, u'/bar?foo=baz')
        self.assertEqual(req.script_root, u'/test')
        self.assertEqual(req.url, u'http://example.com/test/bar?foo=baz')
        self.assertEqual(req.base_url, u'http://example.com/test/bar')
        self.assertEqual(req.url_root, u'http://example.com/test/')
        self.assertEqual(req.host_url, u'http://example.com/')
        self.assertEqual(req.host, 'example.com')
        self.assertEqual(req.scheme, 'http')

        req = Request.from_values(
            '/bar?foo=baz', 'https://example.com/test'
        )
        self.assertEqual(req.scheme, 'https')

        req = Request.from_values(
            '/bar?foo=baz', 'http://example.com/test'
        )
        req.trusted_hosts = ['example.org']
        self.assertRaises(SecurityError, lambda: req.url)
        self.assertRaises(SecurityError, lambda: req.base_url)
        self.assertRaises(SecurityError, lambda: req.url_root)
        self.assertRaises(SecurityError, lambda: req.host_url)
        self.assertRaises(SecurityError, lambda: req.host)

    def test_authorization_mixin(self):
        request = Request.from_values(headers={
            'Authorization': 'Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ=='
        })
        a = request.authorization
        self.assertEqual(a.type, 'basic')
        self.assertEqual(a.username, 'Aladdin')
        self.assertEqual(a.password, 'open sesame')

    def test_stream_only_mixing(self):
        request = PlainRequest.from_values(
            data=b'foo=blub+hehe',
            content_type='application/x-www-form-urlencoded'
        )
        self.assertEqual(list(request.files.items()), [])
        self.assertEqual(list(request.form.items()), [])
        self.assertEqual(request.stream.read(), b'foo=blub+hehe')

    def test_accept_mixin(self):
        request = Request({
            'HTTP_ACCEPT': 'text/xml,application/xml,application/xhtml+xml,'
                           'text/html;q=0.9,text/plain;q=0.8,image/png,'
                           '*/*;q=0.5',
            'HTTP_ACCEPT_CHARSET': 'ISO-8859-1,utf-8;q=0.7,*;q=0.7',
            'HTTP_ACCEPT_ENCODING': 'gzip,deflate',
            'HTTP_ACCEPT_LANGUAGE': 'en-us,en;q=0.5'
        })
        self.assertEqual(request.accept_mimetypes, MIMEAccept([
            ('text/xml', 1), ('image/png', 1), ('application/xml', 1),
            ('application/xhtml+xml', 1), ('text/html', 0.9),
            ('text/plain', 0.8), ('*/*', 0.5)
        ]))
        self.assertEqual(request.accept_charsets, CharsetAccept([
            ('ISO-8859-1', 1), ('utf-8', 0.7), ('*', 0.7)
        ]))
        self.assertEqual(request.accept_encodings, Accept([
            ('gzip', 1), ('deflate', 1)]))
        self.assertEqual(request.accept_languages, LanguageAccept([
            ('en-us', 1), ('en', 0.5)]))

        request = Request({'HTTP_ACCEPT': ''})
        self.assertEqual(request.accept_mimetypes, MIMEAccept())

    def test_etag_request_mixin(self):
        request = Request({
            'HTTP_CACHE_CONTROL':       'no-store, no-cache',
            'HTTP_IF_MATCH':            'w/"foo", bar, "baz"',
            'HTTP_IF_NONE_MATCH':       'w/"foo", bar, "baz"',
            'HTTP_IF_MODIFIED_SINCE':   'Tue, 22 Jan 2008 11:18:44 GMT',
            'HTTP_IF_UNMODIFIED_SINCE': 'Tue, 22 Jan 2008 11:18:44 GMT'
        })
        self.assertTrue(request.cache_control.no_store)
        self.assertTrue(request.cache_control.no_cache)

        for etags in request.if_match, request.if_none_match:
            self.assertTrue(etags('bar'))
            self.assertTrue(etags.contains_raw('w/"foo"'))
            self.assertTrue(etags.contains_weak('foo'))
            self.assertFalse(etags.contains('foo'))

        self.assertEqual(
            request.if_modified_since, datetime(2008, 1, 22, 11, 18, 44)
        )
        self.assertEqual(
            request.if_unmodified_since, datetime(2008, 1, 22, 11, 18, 44)
        )

    def test_user_agent_mixin(self):
        user_agents = [
            ('Mozilla/5.0 (Macintosh; U; Intel Mac OS X; en-US; rv:1.8.1.11) '
             'Gecko/20071127 Firefox/2.0.0.11', 'firefox', 'macos', '2.0.0.11',
             'en-US'),
            ('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; de-DE) '
             'Opera 8.54',
             'opera', 'windows', '8.54', 'de-DE'),
            # ('Mozilla/5.0 (iPhone; U; CPU like Mac OS X; en) AppleWebKit/420'
            # ' (KHTML, like Gecko) Version/3.0 Mobile/1A543a Safari/419.3',
            # 'safari', 'iphone', '3.0', 'en'),
            ('Bot Googlebot/2.1 (http://www.googlebot.com/bot.html)',
             'google', None, '2.1', None),
            ('Mozilla/5.0 (X11; CrOS armv7l 3701.81.0) AppleWebKit/537.31 '
             '(KHTML, like Gecko) Chrome/26.0.1410.57 Safari/537.31',
             'chrome', 'chromeos', '26.0.1410.57', None),
            ('Mozilla/5.0 (Windows NT 6.3; Trident/7.0; .NET4.0E; rv:11.0) '
             'like Gecko',
             'msie', 'windows', '11.0', None),
            ('Mozilla/5.0 (SymbianOS/9.3; Series60/3.2 NokiaE5-00/101.003; '
             'Profile/MIDP-2.1 Configuration/CLDC-1.1 ) AppleWebKit/533.4 '
             '(KHTML, like Gecko) NokiaBrowser/7.3.1.35 '
             'Mobile Safari/533.4 3gpp-gba',
             'safari', 'symbian', '533.4', None)
        ]
        for ua, browser, platform, version, lang in user_agents:
            request = Request({'HTTP_USER_AGENT': ua})
            self.assertEqual(request.user_agent.browser, browser)
            self.assertEqual(request.user_agent.platform, platform)
            self.assertEqual(request.user_agent.version, version)
            self.assertEqual(request.user_agent.language, lang)
            self.assertTrue(bool(request.user_agent))
            self.assertEqual(request.user_agent.to_header(), ua)
            self.assertEqual(str(request.user_agent), ua)

        request = Request({'HTTP_USER_AGENT': 'foo'})
        self.assertFalse(request.user_agent)

    def test_stream_wrapping(self):
        class LowercasingStream(object):

            def __init__(self, stream):
                self._stream = stream

            def read(self, size=-1):
                return self._stream.read(size).lower()

            def readline(self, size=-1):
                return self._stream.readline(size).lower()

        data = b'foo=Hello+World'
        req = Request.from_values(
            '/', method='POST', data=data,
            content_type='application/x-www-form-urlencoded')
        req.stream = LowercasingStream(req.stream)
        self.assertEqual(req.form['foo'], 'hello world')

    def test_data_descriptor_triggers_parsing(self):
        data = b'foo=Hello+World'
        req = Request.from_values(
            '/', method='POST', data=data,
            content_type='application/x-www-form-urlencoded')

        self.assertEqual(req.get_data(parse_form_data=True), b'')
        self.assertEqual(req.form['foo'], u'Hello World')

    def test_get_data_method_parsing_caching_behavior(self):
        data = b'foo=Hello+World'
        req = Request.from_values(
            '/', method='POST', data=data,
            content_type='application/x-www-form-urlencoded')

        # get_data() caches, so form stays available
        self.assertEqual(req.get_data(), data)
        self.assertEqual(req.form['foo'], u'Hello World')
        self.assertEqual(req.get_data(), data)

        # here we access the form data first, caching is bypassed
        req = Request.from_values(
            '/', method='POST', data=data,
            content_type='application/x-www-form-urlencoded')
        self.assertEqual(req.form['foo'], u'Hello World')
        self.assertEqual(req.get_data(), b'')

        # Another case is uncached get data which trashes everything
        req = Request.from_values(
            '/', method='POST', data=data,
            content_type='application/x-www-form-urlencoded')
        self.assertEqual(req.get_data(cache=False), data)
        self.assertEqual(req.get_data(cache=False), b'')
        self.assertEqual(req.form, {})

        # Or we can implicitly start the form parser which is similar to
        # the old .data behavior
        req = Request.from_values(
            '/', method='POST', data=data,
            content_type='application/x-www-form-urlencoded')
        self.assertEqual(req.get_data(parse_form_data=True), b'')
        self.assertEqual(req.form['foo'], u'Hello World')

    def test_common_request_descriptors_mixin(self):
        request = Request.from_values(
            content_type='text/html; charset=utf-8',
            content_length='23',
            headers={
                'Referer':          'http://www.example.com/',
                'Date':             'Sat, 28 Feb 2009 19:04:35 GMT',
                'Max-Forwards':     '10',
                'Pragma':           'no-cache',
                'Content-Encoding': 'gzip',
                'Content-MD5':      '9a3bc6dbc47a70db25b84c6e5867a072'
            }
        )

        self.assertEqual(request.content_type, 'text/html; charset=utf-8')
        self.assertEqual(request.mimetype, 'text/html')
        self.assertEqual(request.mimetype_params, {'charset': 'utf-8'})
        self.assertEqual(request.content_length, 23)
        self.assertEqual(request.referrer, 'http://www.example.com/')
        self.assertEqual(request.date, datetime(2009, 2, 28, 19, 4, 35))
        self.assertEqual(request.max_forwards, 10)
        self.assertTrue('no-cache' in request.pragma)
        self.assertEqual(request.content_encoding, 'gzip')
        self.assertEqual(
            request.content_md5, '9a3bc6dbc47a70db25b84c6e5867a072'
        )

    def test_request_mimetype_always_lowercase(self):
        request = Request.from_values(content_type='APPLICATION/JSON')
        self.assertEqual(request.mimetype, 'application/json')

    def test_shallow_mode(self):
        request = Request({'QUERY_STRING': 'foo=bar'}, shallow=True)
        self.assertEqual(request.args['foo'], 'bar')
        self.assertRaises(RuntimeError, lambda: request.form['foo'])

    def test_form_parsing_failed(self):
        data = (
            b'--blah\r\n'
        )
        data = Request.from_values(
            input_stream=BytesIO(data),
            content_length=len(data),
            content_type='multipart/form-data; boundary=foo',
            method='POST'
        )
        self.assertFalse(data.files)
        self.assertFalse(data.form)

    def test_file_closing(self):
        data = (
            b'--foo\r\n'
            b'Content-Disposition: form-data; name="foo"; '
            b'filename="foo.txt"\r\n'
            b'Content-Type: text/plain; charset=utf-8\r\n\r\n'
            b'file contents, just the contents\r\n'
            b'--foo--'
        )
        req = Request.from_values(
            input_stream=BytesIO(data),
            content_length=len(data),
            content_type='multipart/form-data; boundary=foo',
            method='POST'
        )
        foo = req.files['foo']
        self.assertEqual(foo.mimetype, 'text/plain')
        self.assertEqual(foo.filename, 'foo.txt')

        self.assertFalse(foo.closed)
        req.close()
        self.assertTrue(foo.closed)

    def test_file_closing_with(self):
        data = (
            b'--foo\r\n'
            b'Content-Disposition: form-data; name="foo"; '
            b'filename="foo.txt"\r\n'
            b'Content-Type: text/plain; charset=utf-8\r\n\r\n'
            b'file contents, just the contents\r\n'
            b'--foo--'
        )
        req = Request.from_values(
            input_stream=BytesIO(data),
            content_length=len(data),
            content_type='multipart/form-data; boundary=foo',
            method='POST'
        )
        with req:
            foo = req.files['foo']
            self.assertEqual(foo.mimetype, 'text/plain')
            self.assertEqual(foo.filename, 'foo.txt')

        self.assertTrue(foo.closed)

    def test_url_charset_reflection(self):
        req = Request.from_values()
        req.charset = 'utf-7'
        self.assertEqual(req.url_charset, 'utf-7')

    def test_other_method_payload(self):
        data = b'Hello World'
        req = Request.from_values(
            input_stream=BytesIO(data), method='WHAT_THE_FUCK',
            content_length=len(data), content_type='text/plain'
        )
        self.assertEqual(req.get_data(), data)
        self.assertTrue(isinstance(req.stream, LimitedStream))

    def test_form_data_ordering(self):
        class MyRequest(Request):
            parameter_storage_class = ImmutableOrderedMultiDict

        req = MyRequest.from_values('/?foo=1&bar=0&foo=3')
        self.assertEqual(list(req.args), ['foo', 'bar'])
        self.assertEqual(list(req.args.items(multi=True)), [
            ('foo', '1'),
            ('bar', '0'),
            ('foo', '3')
        ])
        self.assertIsInstance(req.args, ImmutableOrderedMultiDict)
        self.assertIsInstance(req.values, CombinedMultiDict)
        self.assertEqual(req.values['foo'], '1')
        self.assertEqual(req.values.getlist('foo'), ['1', '3'])

    def test_storage_classes(self):
        class MyRequest(Request):
            dict_storage_class = dict
            list_storage_class = list
            parameter_storage_class = dict
        req = MyRequest.from_values('/?foo=baz', headers={
            'Cookie':   'foo=bar'
        })
        self.assertIs(type(req.cookies), dict)
        self.assertEqual(req.cookies, {'foo': 'bar'})
        self.assertIs(type(req.access_route), list)

        self.assertIs(type(req.args), dict)
        self.assertIs(type(req.values), CombinedMultiDict)
        self.assertEqual(req.values['foo'], u'baz')

        req = Request.from_values(headers={
            'Cookie':   'foo=bar'
        })
        self.assertIs(type(req.cookies), ImmutableTypeConversionDict)
        self.assertEqual(req.cookies, {'foo': 'bar'})
        self.assertIs(type(req.access_route), ImmutableList)

        MyRequest.list_storage_class = tuple
        req = MyRequest.from_values()
        self.assertIs(type(req.access_route), tuple)

    def test_modified_url_encoding(self):
        class ModifiedRequest(Request):
            url_charset = 'euc-kr'

        req = ModifiedRequest.from_values(u'/?foo=정상처리'.encode('euc-kr'))
        self.assertEqual(req.args['foo'], u'정상처리')

    def test_request_method_case_sensitivity(self):
        req = Request({'REQUEST_METHOD': 'get'})
        self.assertEqual(req.method, 'GET')


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RequestsTestCase))
    return suite



"""MIME-Type Parser

This module provides basic functions for handling mime-types. It can handle
matching mime-types against a list of media-ranges. See section 14.1 of the
HTTP specification [RFC 2616] for a complete explanation.

   http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.1

Contents:
 - parse_mime_type():   Parses a mime-type into its component parts.
 - parse_media_range(): Media-ranges are mime-types with wild-cards and a 'q'
                          quality parameter.
 - quality():           Determines the quality ('q') of a mime-type when
                          compared against a list of media-ranges.
 - quality_parsed():    Just like quality() except the second parameter must be
                          pre-parsed.
 - best_match():        Choose the mime-type with the highest quality ('q')
                          from a list of candidates.
"""
from copy import copy
from functools import total_ordering


EXACT_MATCH = 2
WILDCARD_MATCH = 1
MISMATCH = 0


class NotAcceptableError(Exception):
    pass


class MimeType(object):
    def __init__(self, type, subtype, params):
        self.type = type
        self.subtype = subtype
        self.params = copy(params)


class MediaRange(MimeType):
    def __init__(self, type, subtype, params=None, q=None):
        super(MediaRange, self).__init__(type, subtype, params)

        if q is None:
            q = 1.0
            if 'q' in self.params:
                q = float(self.params.pop('q'))
        else:
            assert 'q' not in self.params

        if q < 0 or 1 > q:
            q = 1.0
        self.q = q


@total_ordering
class Fitness(object):
    def __init__(self, type_matches, subtype_matches, param_matches):
        self.type_matches = type_matches
        self.subtype_matches = subtype_matches
        self.param_matches = param_matches

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, Fitness):
            return NotImplemented
        if self.type_matches != other.type_matches:
            return False
        if self.subtype_matches != other.subtype_matches:
            return False
        if self.param_matches != other.param_matches:
            return False
        return True

    def __lt__(self, other):
        if other is None:
            return False
        if not isinstance(other, Fitness):
            return NotImplemented
        if self.type_matches < other.type_matches:
            return True
        if self.subtype_matches < other.subtype_matches:
            return True
        if self.param_matches < other.param_matches:
            return True
        return False


def _parse_mime_string(mime_type):
    parts = mime_type.split(';')
    params = {
        key.strip(): value.strip()
        for key, value in (
            param.split('=', 1)
            for param in parts[1:]
        )
    }
    full_type = parts[0].strip()
    # Java URLConnection class sends an Accept header that includes a
    # single '*'. Turn it into a legal wildcard.
    if full_type == '*':
        full_type = '*/*'
    (type, subtype) = full_type.split('/')

    return (type.strip(), subtype.strip(), params)


def parse_mime_type(mime_type):
    """Parses a mime-type into its component parts.

    Carves up a mime-type and returns a tuple of the (type, subtype, params)
    where 'params' is a dictionary of all the parameters for the media range.
    For example, the media range 'application/xhtml;q=0.5' would get parsed
    into:

       ('application', 'xhtml', {'q', '0.5'})
    """

    return MimeType(*_parse_mime_string(mime_type))


def parse_media_range(range):
    """Parse a media-range into its component parts.

    Carves up a media range and returns a tuple of the (type, subtype,
    params) where 'params' is a dictionary of all the parameters for the media
    range.  For example, the media range 'application/*;q=0.5' would get parsed
    into:

       ('application', '*', {'q', '0.5'})

    In addition this function also guarantees that there is a value for 'q'
    in the params dictionary, filling it in with a proper default if
    necessary.
    """
    return MediaRange(*_parse_mime_string(range))


def _compare_types(a, b):
    if a == '*' or b == '*':
        return WILDCARD_MATCH
    elif a == b:
        return EXACT_MATCH
    return MISMATCH


def _fitness(mime_type, media_range):
    type_matches = _compare_types(mime_type.type, media_range.type)
    subtype_matches = _compare_types(mime_type.subtype, media_range.subtype)

    param_matches = sum(
        1 for key in media_range.params
        if key in mime_type.params and
        mime_type.params[key] == media_range.params[key]
    )

    if not type_matches or not subtype_matches:
        return None

    return Fitness(
        type_matches=type_matches,
        subtype_matches=subtype_matches,
        param_matches=param_matches,
    ), media_range.q


def fitness(mime_type, accept):
    if isinstance(mime_type, str):
        mime_type = parse_mime_type(mime_type)
    if isinstance(accept, str):
        accept = [
            parse_media_range(media_range)
            for media_range in accept.split(',')
        ]

    if isinstance(accept, MediaRange):
        value = _fitness(mime_type, accept)
    else:
        if not len(accept):
            raise ValueError()
        # `media_range` is an iterator of media ranges
        # return the fitness of the best match
        value = max(fitness(mime_type, media_range) for media_range in accept)

    if value is None:
        raise NotAcceptableError(mime_type, accept)

    return value
